@@ -26,6 +26,7 @@
  * \file socket.c */
 
 #ifdef HAVE_SYS_SOCKET_H
+#define _GNU_SOURCE /* for struct ucred on glibc >= 2.8 */
 
 #include <osmocom/core/logging.h>
 #include <osmocom/core/select.h>
@@ -171,8 +172,8 @@ static int socket_helper(const struct addrinfo *rp, unsigned int flags)
 
 	sfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
 	if (sfd == -1) {
-		LOGP(DLGLOBAL, LOGL_ERROR,
-			"unable to create socket: %s\n", strerror(errno));
+		LOGP(DLGLOBAL, LOGL_ERROR, "unable to create socket(%d, %d, %d): %s\n",
+		     rp->ai_family, rp->ai_socktype, rp->ai_protocol, strerror(errno));
 		return sfd;
 	}
 
@@ -345,23 +346,23 @@ int osmo_sock_init2(uint16_t family, uint16_t type, uint8_t proto,
 	if (family == AF_UNSPEC) {
 		for (rp = local; rp != NULL; rp = rp->ai_next) {
 			switch (rp->ai_family) {
-				case AF_INET:
-					local_ipv4 = true;
-					break;
-				case AF_INET6:
-					local_ipv6 = true;
-					break;
+			case AF_INET:
+				local_ipv4 = true;
+				break;
+			case AF_INET6:
+				local_ipv6 = true;
+				break;
 			}
 		}
 
 		for (rp = remote; rp != NULL; rp = rp->ai_next) {
 			switch (rp->ai_family) {
-				case AF_INET:
-					remote_ipv4 = true;
-					break;
-				case AF_INET6:
-					remote_ipv6 = true;
-					break;
+			case AF_INET:
+				remote_ipv4 = true;
+				break;
+			case AF_INET6:
+				remote_ipv6 = true;
+				break;
 			}
 		}
 
@@ -616,18 +617,45 @@ int osmo_sock_init_osa(uint16_t type, uint8_t proto,
 
 #ifdef HAVE_LIBSCTP
 
-/* Check whether there's an IPv6 Addr as first option of any addrinfo item in the addrinfo set */
+/* Check whether there's an addrinfo item in the addrinfo set with an IPv4 or IPv6 option */
 static void addrinfo_has_v4v6addr(const struct addrinfo **result, size_t result_count, bool *has_v4, bool *has_v6)
 {
 	size_t host_idx;
+	const struct addrinfo *rp;
 	*has_v4 = false;
 	*has_v6 = false;
 
 	for (host_idx = 0; host_idx < result_count; host_idx++) {
-		if (result[host_idx]->ai_family == AF_INET)
-			*has_v4 = true;
-		else if (result[host_idx]->ai_family == AF_INET6)
-			*has_v6 = true;
+		for (rp = result[host_idx]; rp != NULL; rp = rp->ai_next) {
+			if (result[host_idx]->ai_family == AF_INET)
+				*has_v4 = true;
+			else if (result[host_idx]->ai_family == AF_INET6)
+				*has_v6 = true;
+		}
+	}
+}
+
+/* Check whether there's an addrinfo item in the addrinfo set with only an IPv4 or IPv6 option */
+static void addrinfo_has_v4v6only_addr(const struct addrinfo **result, size_t result_count, bool *has_v4only, bool *has_v6only)
+{
+	size_t host_idx;
+	const struct addrinfo *rp;
+	*has_v4only = false;
+	*has_v6only = false;
+
+	for (host_idx = 0; host_idx < result_count; host_idx++) {
+		bool has_v4 = false;
+		bool has_v6 = false;
+		for (rp = result[host_idx]; rp != NULL; rp = rp->ai_next) {
+			if (rp->ai_family == AF_INET6)
+				has_v6 = true;
+			else
+				has_v4 = true;
+		}
+		if (has_v4 && !has_v6)
+			*has_v4only = true;
+		else if (has_v6 && !has_v4)
+			*has_v6only = true;
 	}
 }
 
@@ -636,13 +664,16 @@ static bool addrinfo_has_in6addr_any(const struct addrinfo **result, size_t resu
 {
 	size_t host_idx;
 	struct in6_addr in6addr_any = IN6ADDR_ANY_INIT;
+	const struct addrinfo *rp;
 
 	for (host_idx = 0; host_idx < result_count; host_idx++) {
-		if (result[host_idx]->ai_family != AF_INET6)
-			continue;
-		if (memcmp(&((struct sockaddr_in6 *)result[host_idx]->ai_addr)->sin6_addr,
-			   &in6addr_any, sizeof(in6addr_any)) == 0)
-			return true;
+		for (rp = result[host_idx]; rp != NULL; rp = rp->ai_next) {
+			if (rp->ai_family != AF_INET6)
+				continue;
+			if (memcmp(&((struct sockaddr_in6 *)rp->ai_addr)->sin6_addr,
+				&in6addr_any, sizeof(in6addr_any)) == 0)
+				return true;
+		}
 	}
 	return false;
 }
@@ -676,22 +707,28 @@ static int addrinfo_to_sockaddr(uint16_t family, const struct addrinfo **result,
 	for (host_idx = 0; host_idx < host_cont; host_idx++) {
 		/* Addresses are ordered based on RFC 3484, see man getaddrinfo */
 		for (rp = result[host_idx]; rp != NULL; rp = rp->ai_next) {
-			if (family != AF_UNSPEC && rp->ai_family != family)
-				continue;
-			if (offset + rp->ai_addrlen > addrs_buf_len) {
-				LOGP(DLGLOBAL, LOGL_ERROR, "Output buffer to small: %zu\n",
-				     addrs_buf_len);
-				return -ENOSPC;
+			if (family == AF_UNSPEC || rp->ai_family == family)
+				break;
+		}
+		if (!rp && family == AF_INET6) {
+			/* See if we can find an AF_INET addr for the AF_INET6 socket instead: */
+			for (rp = result[host_idx]; rp != NULL; rp = rp->ai_next) {
+				if (rp->ai_family == AF_INET)
+					break;
 			}
-			memcpy(addrs_buf + offset, rp->ai_addr, rp->ai_addrlen);
-			offset += rp->ai_addrlen;
-			break;
 		}
 		if (!rp) { /* No addr could be bound for this host! */
 			LOGP(DLGLOBAL, LOGL_ERROR, "No suitable remote address found for host: %s\n",
 			     hosts[host_idx]);
 			return -ENODEV;
 		}
+		if (offset + rp->ai_addrlen > addrs_buf_len) {
+			LOGP(DLGLOBAL, LOGL_ERROR, "Output buffer to small: %zu\n",
+				addrs_buf_len);
+			return -ENOSPC;
+		}
+		memcpy(addrs_buf + offset, rp->ai_addr, rp->ai_addrlen);
+		offset += rp->ai_addrlen;
 	}
 	return 0;
 }
@@ -827,15 +864,13 @@ int osmo_sock_init2_multiaddr2(uint16_t family, uint16_t type, uint8_t proto,
 	struct addrinfo *res_loc[OSMO_SOCK_MAX_ADDRS], *res_rem[OSMO_SOCK_MAX_ADDRS];
 	int sfd = -1, rc, on = 1;
 	unsigned int i;
-	bool loc_has_v4addr, rem_has_v4addr;
-	bool loc_has_v6addr, rem_has_v6addr;
+	bool loc_has_v4addr = false, loc_has_v6addr = false;
+	bool rem_has_v4addr = false, rem_has_v6addr = false;
+	bool loc_has_v4only_addr, rem_has_v4only_addr;
+	bool loc_has_v6only_addr, rem_has_v6only_addr;
 	struct sockaddr_in6 addrs_buf[OSMO_SOCK_MAX_ADDRS];
 	char strbuf[512];
 
-	/* updated later in case of AF_UNSPEC */
-	loc_has_v4addr = rem_has_v4addr = (family == AF_INET);
-	loc_has_v6addr = rem_has_v6addr = (family == AF_INET6);
-
 	/* TODO: So far this function is only aimed for SCTP, but could be
 	   reused in the future for other protocols with multi-addr support */
 	if (proto != IPPROTO_SCTP)
@@ -862,10 +897,17 @@ int osmo_sock_init2_multiaddr2(uint16_t family, uint16_t type, uint8_t proto,
 					   local_hosts_cnt, local_port, true);
 		if (rc < 0)
 			return -EINVAL;
-		/* Figure out if there's any IPV4 or IPv6 addr in the set */
-		if (family == AF_UNSPEC)
-			addrinfo_has_v4v6addr((const struct addrinfo **)res_loc, local_hosts_cnt,
-					      &loc_has_v4addr, &loc_has_v6addr);
+		/* Figure out if there's any IPv4 or IPv6 entry in the result set */
+		addrinfo_has_v4v6addr((const struct addrinfo **)res_loc, local_hosts_cnt,
+				       &loc_has_v4addr, &loc_has_v6addr);
+		/* Figure out if there's any IPv4-only or IPv6-only addr in the result set */
+		addrinfo_has_v4v6only_addr((const struct addrinfo **)res_loc, local_hosts_cnt,
+					   &loc_has_v4only_addr, &loc_has_v6only_addr);
+		if (family == AF_INET && loc_has_v6only_addr) {
+			LOGP(DLGLOBAL, LOGL_ERROR, "Cannot bind an IPv6 address to an AF_INET socket\n");
+			rc = -EINVAL;
+			goto ret_freeaddrinfo_loc;
+		}
 	}
 	/* figure out remote side of socket */
 	if (flags & OSMO_SOCK_F_CONNECT) {
@@ -875,31 +917,96 @@ int osmo_sock_init2_multiaddr2(uint16_t family, uint16_t type, uint8_t proto,
 			rc = -EINVAL;
 			goto ret_freeaddrinfo_loc;
 		}
-		/* Figure out if there's any IPv4 or IPv6  addr in the set */
-		if (family == AF_UNSPEC)
-			addrinfo_has_v4v6addr((const struct addrinfo **)res_rem, remote_hosts_cnt,
-					      &rem_has_v4addr, &rem_has_v6addr);
+		/* Figure out if there's any IPv4 or IPv6 entry in the result set */
+		addrinfo_has_v4v6addr((const struct addrinfo **)res_rem, remote_hosts_cnt,
+				       &rem_has_v4addr, &rem_has_v6addr);
+		/* Figure out if there's any IPv4-only or IPv6-only addr in the result set */
+		addrinfo_has_v4v6only_addr((const struct addrinfo **)res_rem, remote_hosts_cnt,
+					   &rem_has_v4only_addr, &rem_has_v6only_addr);
+		if (family == AF_INET && rem_has_v6only_addr) {
+			LOGP(DLGLOBAL, LOGL_ERROR, "Cannot connect to an IPv6 address in an AF_INET socket\n");
+			rc = -EINVAL;
+			goto ret_freeaddrinfo;
+		}
 	}
 
-	if (((flags & OSMO_SOCK_F_BIND) && (flags & OSMO_SOCK_F_CONNECT)) &&
-	    !addrinfo_has_in6addr_any((const struct addrinfo **)res_loc, local_hosts_cnt) &&
+	/* Find out the socket family now if not established by caller:
+	 * Both are checked here through "or" here to account for "bind flag set,
+	 * connect flag not set" and viceversa. */
+	if (family == AF_UNSPEC) {
+		if (!loc_has_v6addr && !rem_has_v6addr)
+			family = AF_INET;
+		else
+			family = AF_INET6;
+	}
+
+	/* if both sets are used, make sure there's at least 1 address of the
+	 * same type on each set so that SCTP INIT/INIT-ACK can work. */
+	if (family == AF_INET6 && ((flags & OSMO_SOCK_F_BIND) && (flags & OSMO_SOCK_F_CONNECT)) &&
 	    (loc_has_v4addr != rem_has_v4addr || loc_has_v6addr != rem_has_v6addr)) {
-		LOGP(DLGLOBAL, LOGL_ERROR, "Invalid v4 vs v6 in local vs remote addresses: "
-		     "local:%s%s remote:%s%s\n",
-		     loc_has_v4addr ? " v4" : "", loc_has_v6addr ? " v6" : "",
-		     rem_has_v4addr ? " v4" : "", rem_has_v6addr ? " v6" : ""
-		     );
-		rc = -EINVAL;
-		goto ret_freeaddrinfo;
+		if (!addrinfo_has_in6addr_any((const struct addrinfo **)res_loc, local_hosts_cnt)) {
+			LOGP(DLGLOBAL, LOGL_ERROR, "Invalid v4 vs v6 in local vs remote addresses: "
+			     "local:%s%s remote:%s%s\n",
+			     loc_has_v4addr ? " v4" : "", loc_has_v6addr ? " v6" : "",
+			     rem_has_v4addr ? " v4" : "", rem_has_v6addr ? " v6" : "");
+			rc = -EINVAL;
+			goto ret_freeaddrinfo;
+		}
 	}
 
-	sfd = socket_helper_multiaddr(loc_has_v6addr ? AF_INET6 : AF_INET,
-				      type, proto, flags);
+	sfd = socket_helper_multiaddr(family, type, proto, flags);
 	if (sfd < 0) {
 		rc = sfd;
 		goto ret_freeaddrinfo;
 	}
 
+	if (pars) {
+		if (pars->sctp.sockopt_auth_supported.set) {
+			/* RFC 5061 4.2.7: ASCONF also requires AUTH feature. */
+			rc = setsockopt_sctp_auth_supported(sfd, pars->sctp.sockopt_auth_supported.value);
+			if (rc < 0) {
+				int err = errno;
+				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
+				LOGP(DLGLOBAL, LOGL_ERROR,
+				     "cannot setsockopt(SCTP_AUTH_SUPPORTED) socket: %s:%u: %s\n",
+				     strbuf, local_port, strerror(err));
+				if (pars->sctp.sockopt_auth_supported.abort_on_failure)
+					goto ret_close;
+				/* do not fail, some features such as Peer Primary Address won't be available
+				 * unless configured system-wide through sysctl */
+			}
+		}
+
+		if (pars->sctp.sockopt_asconf_supported.set) {
+			rc = setsockopt_sctp_asconf_supported(sfd, pars->sctp.sockopt_asconf_supported.value);
+			if (rc < 0) {
+				int err = errno;
+				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
+				LOGP(DLGLOBAL, LOGL_ERROR,
+				     "cannot setsockopt(SCTP_ASCONF_SUPPORTED) socket: %s:%u: %s\n",
+				     strbuf, local_port, strerror(err));
+				if (pars->sctp.sockopt_asconf_supported.abort_on_failure)
+					goto ret_close;
+				/* do not fail, some features such as Peer Primary Address won't be available
+				 * unless configured system-wide through sysctl */
+			}
+		}
+
+		if (pars->sctp.sockopt_initmsg.set) {
+			rc = setsockopt_sctp_initmsg(sfd, pars);
+			if (rc < 0) {
+				int err = errno;
+				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
+				LOGP(DLGLOBAL, LOGL_ERROR,
+				     "cannot setsockopt(SCTP_INITMSG) socket: %s:%u: %s\n",
+				     strbuf, local_port, strerror(err));
+				if (pars->sctp.sockopt_initmsg.abort_on_failure)
+					goto ret_close;
+				/* do not fail, some parameters will be left as the global default */
+			}
+		}
+	}
+
 	if (flags & OSMO_SOCK_F_BIND) {
 		/* Since so far we only allow IPPROTO_SCTP in this function,
 		   no need to check below for "proto != IPPROTO_UDP || flags & OSMO_SOCK_F_UDP_REUSEADDR" */
@@ -916,51 +1023,6 @@ int osmo_sock_init2_multiaddr2(uint16_t family, uint16_t type, uint8_t proto,
 			goto ret_close;
 		}
 
-		if (pars && pars->sctp.sockopt_auth_supported.set) {
-			/* RFC 5061 4.2.7: ASCONF also requires AUTH feature. */
-			rc = setsockopt_sctp_auth_supported(sfd, pars->sctp.sockopt_auth_supported.value);
-			if (rc < 0) {
-				int err = errno;
-				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
-				LOGP(DLGLOBAL, LOGL_ERROR,
-				     "cannot setsockopt(SCTP_AUTH_SUPPORTED) socket: %s:%u: %s\n",
-				     strbuf, local_port, strerror(err));
-				if (pars->sctp.sockopt_auth_supported.abort_on_failure)
-					goto ret_close;
-				/* do not fail, some features such as Peer Primary Address won't be available
-				 * unless configured system-wide through sysctl */
-			}
-		}
-
-		if (pars && pars->sctp.sockopt_asconf_supported.set) {
-			rc = setsockopt_sctp_asconf_supported(sfd, pars->sctp.sockopt_asconf_supported.value);
-			if (rc < 0) {
-				int err = errno;
-				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
-				LOGP(DLGLOBAL, LOGL_ERROR,
-				     "cannot setsockopt(SCTP_ASCONF_SUPPORTED) socket: %s:%u: %s\n",
-				     strbuf, local_port, strerror(err));
-				if (pars->sctp.sockopt_asconf_supported.abort_on_failure)
-					goto ret_close;
-				/* do not fail, some features such as Peer Primary Address won't be available
-				 * unless configured system-wide through sysctl */
-			}
-		}
-
-		if (pars && pars->sctp.sockopt_initmsg.set) {
-			rc = setsockopt_sctp_initmsg(sfd, pars);
-			if (rc < 0) {
-				int err = errno;
-				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
-				LOGP(DLGLOBAL, LOGL_ERROR,
-				     "cannot setsockopt(SCTP_INITMSG) socket: %s:%u: %s\n",
-				     strbuf, local_port, strerror(err));
-				if (pars->sctp.sockopt_initmsg.abort_on_failure)
-					goto ret_close;
-				/* do not fail, some parameters will be left as the global default */
-			}
-		}
-
 		/* Build array of addresses taking first entry for each host.
 		   TODO: Ideally we should use backtracking storing last used
 		   indexes and trying next combination if connect() fails .*/
@@ -1257,6 +1319,160 @@ int osmo_sock_init_sa(struct sockaddr *ss, uint16_t type,
 	return osmo_sock_init(ss->sa_family, type, proto, host, port, flags);
 }
 
+#ifdef HAVE_LIBSCTP
+/*! Add addresses to the multi-address (SCTP) socket active binding set
+ *  \param[in] sfd The multi-address (SCTP) socket
+ *  \param[in] addrs array of char pointers (strings), each containing local host name or IP address in string form
+ *  \param[in] addrs_cnt length of addrs_hosts (in items)
+ *  \returns 0 on success; negative on error
+ *
+ * This function only supports SCTP sockets so far, and hence it should be
+ * called only on socket file descriptions referencing that kind of sockets.
+ */
+int osmo_sock_multiaddr_add_local_addr(int sfd, const char **addrs, size_t addrs_cnt)
+{
+	struct osmo_sockaddr osa;
+	socklen_t slen = sizeof(osa);
+	uint16_t sfd_family;
+	uint16_t type = SOCK_STREAM ;/* Fixme: we assume fd is SOCK_STREAM */
+	uint8_t proto = IPPROTO_SCTP; /* Fixme: we assume fd is IPPROTO_SCTP */
+	struct addrinfo *res[OSMO_SOCK_MAX_ADDRS];
+	uint16_t port;
+	struct sockaddr_in6 addrs_buf[OSMO_SOCK_MAX_ADDRS];
+	unsigned int i;
+	int rc;
+	bool res_has_v4addr = false, res_has_v6addr = false;
+
+	rc = getsockname(sfd, &osa.u.sa, &slen);
+	if (rc < 0)
+		return rc; /* TODO: log error? */
+	sfd_family = osa.u.sa.sa_family;
+	port = osmo_sockaddr_port(&osa.u.sa);
+
+	if (sfd_family != AF_INET && sfd_family != AF_INET6)
+		return -EINVAL;
+
+	rc = addrinfo_helper_multi(res, AF_UNSPEC, type, proto, addrs,
+				   addrs_cnt, port, true);
+	if (rc < 0)
+		return -EINVAL;
+
+	addrinfo_has_v4v6addr((const struct addrinfo **)res, addrs_cnt,
+			      &res_has_v4addr, &res_has_v6addr);
+	if (sfd_family == AF_INET && !res_has_v4addr) {
+		rc = -EINVAL;
+		goto ret_free;
+	}
+
+	uint16_t new_addr_family;
+	if (sfd_family == AF_INET)
+		new_addr_family = AF_INET;
+	else if (sfd_family == AF_INET6 && !res_has_v4addr)
+		new_addr_family = AF_INET6;
+	else
+		new_addr_family = AF_UNSPEC;
+	rc = addrinfo_to_sockaddr(new_addr_family, (const struct addrinfo **)res,
+				  addrs, addrs_cnt,
+				  (uint8_t *)addrs_buf, sizeof(addrs_buf));
+	if (rc < 0) {
+		rc = -ENODEV;
+		goto ret_free;
+	}
+
+	rc = sctp_bindx(sfd, (struct sockaddr *)addrs_buf, addrs_cnt, SCTP_BINDX_ADD_ADDR);
+	if (rc == -1) {
+		int err = errno;
+		char strbuf[512];
+		multiaddr_snprintf(strbuf, sizeof(strbuf), addrs, addrs_cnt);
+		LOGP(DLGLOBAL, LOGL_NOTICE, "Unable to bind socket to new addresses: %s:%u: %s\n",
+			strbuf, port, strerror(err));
+		rc = -ENODEV;
+		goto ret_free;
+	}
+
+ret_free:
+	for (i = 0; i < addrs_cnt; i++)
+		freeaddrinfo(res[i]);
+	return rc;
+}
+
+/*! Remove addresses from the multi-address (SCTP) socket active binding set
+ *  \param[in] sfd The multi-address (SCTP) socket
+ *  \param[in] addrs array of char pointers (strings), each containing local host name or IP address in string form
+ *  \param[in] addrs_cnt length of addrs_hosts (in items)
+ *  \returns 0 on success; negative on error
+ *
+ * This function only supports SCTP sockets so far, and hence it should be
+ * called only on socket file descriptions referencing that kind of sockets.
+ */
+int osmo_sock_multiaddr_del_local_addr(int sfd, const char **addrs, size_t addrs_cnt)
+{
+	struct osmo_sockaddr osa;
+	socklen_t slen = sizeof(osa);
+	uint16_t sfd_family;
+	uint16_t type = SOCK_STREAM ;/* Fixme: we assume fd is SOCK_STREAM */
+	uint8_t proto = IPPROTO_SCTP; /* Fixme: we assume fd is IPPROTO_SCTP */
+	struct addrinfo *res[OSMO_SOCK_MAX_ADDRS];
+	uint16_t port;
+	struct sockaddr_in6 addrs_buf[OSMO_SOCK_MAX_ADDRS];
+	unsigned int i;
+	int rc;
+	bool res_has_v4addr = false, res_has_v6addr = false;
+
+	rc = getsockname(sfd, &osa.u.sa, &slen);
+	if (rc < 0)
+		return rc; /* TODO: log error? */
+	sfd_family = osa.u.sa.sa_family;
+	port = osmo_sockaddr_port(&osa.u.sa);
+
+	if (sfd_family != AF_INET && sfd_family != AF_INET6)
+		return -EINVAL;
+
+	rc = addrinfo_helper_multi(res, AF_UNSPEC, type, proto, addrs,
+				   addrs_cnt, port, true);
+	if (rc < 0)
+		return -EINVAL;
+
+	addrinfo_has_v4v6addr((const struct addrinfo **)res, addrs_cnt,
+			      &res_has_v4addr, &res_has_v6addr);
+	if (sfd_family == AF_INET && !res_has_v4addr) {
+		rc = -EINVAL;
+		goto ret_free;
+	}
+
+	uint16_t del_addr_family;
+	if (sfd_family == AF_INET)
+		del_addr_family = AF_INET;
+	else if (sfd_family == AF_INET6 && !res_has_v4addr)
+		del_addr_family = AF_INET6;
+	else
+		del_addr_family = AF_UNSPEC;
+	rc = addrinfo_to_sockaddr(del_addr_family, (const struct addrinfo **)res,
+				  addrs, addrs_cnt,
+				  (uint8_t *)addrs_buf, sizeof(addrs_buf));
+	if (rc < 0) {
+		rc = -ENODEV;
+		goto ret_free;
+	}
+
+	rc = sctp_bindx(sfd, (struct sockaddr *)addrs_buf, addrs_cnt, SCTP_BINDX_REM_ADDR);
+	if (rc == -1) {
+		int err = errno;
+		char strbuf[512];
+		multiaddr_snprintf(strbuf, sizeof(strbuf), addrs, addrs_cnt);
+		LOGP(DLGLOBAL, LOGL_NOTICE, "Unable to unbind socket from addresses: %s:%u: %s\n",
+			strbuf, port, strerror(err));
+		rc = -ENODEV;
+		goto ret_free;
+	}
+
+ret_free:
+	for (i = 0; i < addrs_cnt; i++)
+		freeaddrinfo(res[i]);
+	return rc;
+}
+#endif /* HAVE_LIBSCTP */
+
 static int sockaddr_equal(const struct sockaddr *a,
 			  const struct sockaddr *b, unsigned int len)
 {
@@ -1508,6 +1724,27 @@ int osmo_sockaddr_netmask_to_prefixlen(const struct osmo_sockaddr *netmask)
 	}
 }
 
+/*! Convert an IP address string (and port number) into a 'struct osmo_sockaddr'.
+ *  \param[out] osa_out caller-allocated osmo_sockaddr storage
+ *  \param[in] ipstr IP[v4,v6] address in string format
+ *  \param[in] port port number (host byte order)
+ *  \returns 0 on success; negative on error. */
+int osmo_sockaddr_from_str_and_uint(struct osmo_sockaddr *osa_out, const char *ipstr, uint16_t port)
+{
+	struct addrinfo *ai = addrinfo_helper(AF_UNSPEC, 0, 0, ipstr, port, true);
+
+	if (!ai)
+		return -EIO;
+
+	if (ai->ai_addrlen > sizeof(*osa_out))
+		return -ENOSPC;
+
+	memcpy(&osa_out->u.sa, ai->ai_addr, ai->ai_addrlen);
+	freeaddrinfo(ai);
+
+	return 0;
+}
+
 /*! Initialize a unix domain socket (including bind/connect)
  *  \param[in] type Socket type like SOCK_DGRAM, SOCK_STREAM
  *  \param[in] proto Protocol like IPPROTO_TCP, IPPROTO_UDP
@@ -1632,6 +1869,97 @@ int osmo_sock_get_ip_and_port(int fd, char *ip, size_t ip_len, char *port, size_
 	return 0;
 }
 
+#ifdef HAVE_LIBSCTP
+/*! Get multiple IP addresses and/or port number on socket in separate string buffers
+ *  \param[in] fd file descriptor of socket.
+ *  \param[out] ip_proto IPPROTO of the socket, eg: IPPROTO_SCTP.
+ *  \param[out] ip Pointer to memory holding consecutive buffers of size ip_len.
+ *  \param[out] ip_cnt length ip array pointer. on return it contains the number of addresses found.
+ *  \param[in] ip_len length of each of the string buffer in the the ip array.
+ *  \param[out] port number (will be filled in when not NULL).
+ *  \param[in] port_len length of the port buffer.
+ *  \param[in] local (true) or remote (false) name will get looked at.
+ *  \returns 0 on success; negative otherwise.
+ *
+ * Upon return, ip_cnt can be set to a higher value than the one set by the
+ * caller. This can be used by the caller to find out the required array length
+ * and then obtaining by calling the function twice. Only up to ip_cnt addresses
+ * are filed in, as per the value provided by the caller.
+ *
+ * Usage example retrieving all (up to OSMO_SOCK_MAX_ADDRS, 32) bound IP addresses and bound port:
+ * char hostbuf[OSMO_SOCK_MAX_ADDRS][INET6_ADDRSTRLEN];
+ * size_t num_hostbuf = ARRAY_SIZE(hostbuf);
+ * char portbuf[6];
+ * rc = osmo_sock_multiaddr_get_ip_and_port(fd, IPPROTO_SCTP, &hostbuf[0][0], &num_hostbuf,
+ *					    sizeof(hostbuf[0]), portbuf, sizeof(portbuf), true);
+ * if (rc < 0)
+ *	goto error;
+ * if (num_hostbuf > ARRAY_SIZE(hostbuf))
+ *	goto not_enough_buffers;
+ */
+int osmo_sock_multiaddr_get_ip_and_port(int fd, int ip_proto, char *ip, size_t *ip_cnt, size_t ip_len,
+					char *port, size_t port_len, bool local)
+{
+	struct sockaddr *addrs = NULL;
+	unsigned int n_addrs, i;
+	void *addr_buf;
+	int rc;
+
+	switch (ip_proto) {
+	case IPPROTO_SCTP:
+		break; /* continue below */
+	default:
+		if (*ip_cnt == 0) {
+			*ip_cnt = 1;
+			return 0;
+		}
+		*ip_cnt = 1;
+		return osmo_sock_get_ip_and_port(fd, ip, ip_len, port, port_len, local);
+	}
+
+	rc = local ? sctp_getladdrs(fd, 0, &addrs) : sctp_getpaddrs(fd, 0, &addrs);
+	if (rc < 0)
+		return rc;
+	if (rc == 0)
+		return -ENOTCONN;
+
+	n_addrs = rc;
+	addr_buf = (void *)addrs;
+	for (i = 0; i < n_addrs; i++) {
+		struct sockaddr *sa_addr = (struct sockaddr *)addr_buf;
+		size_t addrlen;
+
+		if (i >= *ip_cnt)
+			break;
+
+		switch (sa_addr->sa_family) {
+		case AF_INET:
+			addrlen = sizeof(struct sockaddr_in);
+			break;
+		case AF_INET6:
+			addrlen = sizeof(struct sockaddr_in6);
+			break;
+		default:
+			rc = -EINVAL;
+			goto free_addrs_ret;
+		}
+
+		rc = getnameinfo(sa_addr, addrlen, &ip[i * ip_len], ip_len,
+				 port, port_len,
+				 NI_NUMERICHOST | NI_NUMERICSERV);
+		if (rc < 0)
+			goto free_addrs_ret;
+		addr_buf += addrlen;
+	}
+
+	*ip_cnt = n_addrs;
+	rc = 0;
+free_addrs_ret:
+	local ? sctp_freeladdrs(addrs) : sctp_freepaddrs(addrs);
+	return rc;
+}
+#endif
+
 /*! Get local IP address on socket
  *  \param[in] fd file descriptor of socket
  *  \param[out] ip IP address (will be filled in)
@@ -1693,6 +2021,114 @@ char *osmo_sock_get_name(const void *ctx, int fd)
 	return talloc_asprintf(ctx, "(%s)", str);
 }
 
+#ifdef HAVE_LIBSCTP
+/*! Format multiple IP addresses and/or port number into a combined string buffer
+ *  \param[out] str  Destination string buffer.
+ *  \param[in] str_len  sizeof(str), usually OSMO_SOCK_MULTIADDR_PEER_STR_MAXLEN.
+ *  \param[out] ip Pointer to memory holding ip_cnt consecutive buffers of size ip_len.
+ *  \param[out] ip_cnt length ip array pointer. on return it contains the number of addresses found.
+ *  \param[in] ip_len length of each of the string buffer in the the ip array.
+ *  \param[out] port number (will be printed in when not NULL).
+ *  \return String length as returned by snprintf(), or negative on error.
+ *
+ * This API expects an ip array as the one filled in by
+ * osmo_sock_multiaddr_get_ip_and_port(), and hence it's a good companion for
+ * that API.
+ */
+int osmo_multiaddr_ip_and_port_snprintf(char *str, size_t str_len,
+					const char *ip, size_t ip_cnt, size_t ip_len,
+					const char *portbuf)
+{
+	struct osmo_strbuf sb = { .buf = str, .len = str_len };
+	bool is_v6 = false;
+	unsigned int i;
+
+	if (ip_cnt == 0) {
+		OSMO_STRBUF_PRINTF(sb, "NULL:%s", portbuf);
+		return sb.chars_needed;
+	}
+	if (ip_cnt > 1)
+		OSMO_STRBUF_PRINTF(sb, "(");
+	else if ((is_v6 = !!strchr(&ip[0], ':'))) /* IPv6, add [] to separate from port. */
+		OSMO_STRBUF_PRINTF(sb, "[");
+
+	for (i = 0; i < ip_cnt - 1; i++)
+		OSMO_STRBUF_PRINTF(sb, "%s|", &ip[i * ip_len]);
+	OSMO_STRBUF_PRINTF(sb, "%s", &ip[i * ip_len]);
+
+	if (ip_cnt > 1)
+		OSMO_STRBUF_PRINTF(sb, ")");
+	else if (is_v6)
+		OSMO_STRBUF_PRINTF(sb, "]");
+	if (portbuf)
+		OSMO_STRBUF_PRINTF(sb, ":%s", portbuf);
+
+	return sb.chars_needed;
+}
+
+/*! Get address/port information on socket in provided string buffer, like "r=1.2.3.4:5<->l=6.7.8.9:10".
+ * This does not include braces like osmo_sock_get_name().
+ *  \param[out] str  Destination string buffer.
+ *  \param[in] str_len  sizeof(str), usually OSMO_SOCK_MULTIADDR_NAME_MAXLEN.
+ *  \param[in] fd  File descriptor of socket.
+ *  \param[in] fd IPPROTO of the socket, eg: IPPROTO_SCTP.
+ *  \return String length as returned by snprintf(), or negative on error.
+ */
+int osmo_sock_multiaddr_get_name_buf(char *str, size_t str_len, int fd, int sk_proto)
+{
+	char hostbuf[OSMO_SOCK_MAX_ADDRS][INET6_ADDRSTRLEN];
+	size_t num_hostbuf = ARRAY_SIZE(hostbuf);
+	char portbuf[6];
+	struct osmo_strbuf sb = { .buf = str, .len = str_len };
+
+	if (fd < 0) {
+		osmo_strlcpy(str, "<error-bad-fd>", str_len);
+		return sb.chars_needed;
+	}
+
+	switch (sk_proto) {
+	case IPPROTO_SCTP:
+		break; /* continue below */
+	default:
+		return osmo_sock_get_name_buf(str, str_len, fd);
+	}
+
+	/* get remote */
+	OSMO_STRBUF_PRINTF(sb, "r=");
+	if (osmo_sock_multiaddr_get_ip_and_port(fd, sk_proto, &hostbuf[0][0], &num_hostbuf,
+						sizeof(hostbuf[0]), portbuf, sizeof(portbuf), false) != 0) {
+		OSMO_STRBUF_PRINTF(sb, "NULL");
+	} else {
+		const bool need_more_bufs = num_hostbuf > ARRAY_SIZE(hostbuf);
+		if (need_more_bufs)
+			num_hostbuf = ARRAY_SIZE(hostbuf);
+		OSMO_STRBUF_APPEND(sb, osmo_multiaddr_ip_and_port_snprintf,
+				   &hostbuf[0][0], num_hostbuf, sizeof(hostbuf[0]), portbuf);
+		if (need_more_bufs)
+			OSMO_STRBUF_PRINTF(sb, "<need-more-bufs!>");
+	}
+
+	OSMO_STRBUF_PRINTF(sb, "<->l=");
+
+	/* get local */
+	num_hostbuf = ARRAY_SIZE(hostbuf);
+	if (osmo_sock_multiaddr_get_ip_and_port(fd, sk_proto, &hostbuf[0][0], &num_hostbuf,
+						sizeof(hostbuf[0]), portbuf, sizeof(portbuf), true) != 0) {
+		OSMO_STRBUF_PRINTF(sb, "NULL");
+	} else {
+		const bool need_more_bufs = num_hostbuf > ARRAY_SIZE(hostbuf);
+		if (need_more_bufs)
+			num_hostbuf = ARRAY_SIZE(hostbuf);
+		OSMO_STRBUF_APPEND(sb, osmo_multiaddr_ip_and_port_snprintf,
+				   &hostbuf[0][0], num_hostbuf, sizeof(hostbuf[0]), portbuf);
+		if (need_more_bufs)
+			OSMO_STRBUF_PRINTF(sb, "<need-more-bufs!>");
+	}
+
+	return sb.chars_needed;
+}
+#endif
+
 /*! Get address/port information on socket in provided string buffer, like "r=1.2.3.4:5<->l=6.7.8.9:10".
  * This does not include braces like osmo_sock_get_name().
  *  \param[out] str  Destination string buffer.
@@ -1702,8 +2138,10 @@ char *osmo_sock_get_name(const void *ctx, int fd)
  */
 int osmo_sock_get_name_buf(char *str, size_t str_len, int fd)
 {
-	char hostbuf_l[INET6_ADDRSTRLEN], hostbuf_r[INET6_ADDRSTRLEN];
-	char portbuf_l[6], portbuf_r[6];
+	struct osmo_strbuf sb = { .buf = str, .len = str_len };
+	struct osmo_sockaddr osa;
+	struct sockaddr_un *sun;
+	socklen_t len;
 	int rc;
 
 	if (fd < 0) {
@@ -1711,17 +2149,78 @@ int osmo_sock_get_name_buf(char *str, size_t str_len, int fd)
 		return -EBADF;
 	}
 
-	/* get local */
-	if ((rc = osmo_sock_get_ip_and_port(fd, hostbuf_l, sizeof(hostbuf_l), portbuf_l, sizeof(portbuf_l), true))) {
+
+	len = sizeof(osa.u.sas);
+	rc = getsockname(fd, &osa.u.sa, &len);
+	if (rc < 0) {
 		osmo_strlcpy(str, "<error-in-getsockname>", str_len);
 		return rc;
 	}
 
-	/* get remote */
-	if (osmo_sock_get_ip_and_port(fd, hostbuf_r, sizeof(hostbuf_r), portbuf_r, sizeof(portbuf_r), false) != 0)
-		return snprintf(str, str_len, "r=NULL<->l=%s:%s", hostbuf_l, portbuf_l);
+	switch (osa.u.sa.sa_family) {
+	case AF_INET:
+	case AF_INET6:
+	{
+		char hostbuf_l[INET6_ADDRSTRLEN], hostbuf_r[INET6_ADDRSTRLEN];
+		char portbuf_l[6], portbuf_r[6];
 
-	return snprintf(str, str_len, "r=%s:%s<->l=%s:%s", hostbuf_r, portbuf_r, hostbuf_l, portbuf_l);
+		len = sizeof(osa.u.sas);
+		rc = getnameinfo(&osa.u.sa, len, hostbuf_l, sizeof(hostbuf_l),
+				 portbuf_l, sizeof(portbuf_l),
+				 NI_NUMERICHOST | NI_NUMERICSERV);
+		if (rc < 0) {
+			osmo_strlcpy(str, "<error-in-getnameinfo>", str_len);
+			return rc;
+		}
+		/* Now attempt to get remote: */
+		len = sizeof(osa.u.sas);
+		rc = getpeername(fd, &osa.u.sa, &len);
+		if (rc < 0) {
+			OSMO_STRBUF_PRINTF(sb, "r=NULL<->l=%s:%s", hostbuf_l, portbuf_l);
+			return sb.chars_needed;
+		}
+		len = sizeof(osa.u.sas);
+		rc = getnameinfo(&osa.u.sa, len, hostbuf_r, sizeof(hostbuf_r),
+				 portbuf_r, sizeof(portbuf_r),
+				 NI_NUMERICHOST | NI_NUMERICSERV);
+		if (rc < 0) {
+			OSMO_STRBUF_PRINTF(sb, "r=NULL<->l=%s:%s", hostbuf_l, portbuf_l);
+			return sb.chars_needed;
+		}
+		OSMO_STRBUF_PRINTF(sb, "r=%s:%s<->l=%s:%s", hostbuf_r, portbuf_r, hostbuf_l, portbuf_l);
+		return sb.chars_needed;
+	}
+	case AF_UNIX:
+	{
+		unsigned long long remote_pid;
+		bool have_remote_pid;
+#if defined(SO_PEERCRED)
+		struct ucred ucred;
+		len = sizeof(struct ucred);
+		if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &ucred, &len) == -1) {
+			have_remote_pid = false;
+		} else {
+			have_remote_pid = true;
+			remote_pid = (unsigned long long)ucred.pid;
+		}
+#else
+		#pragma message "SO_PEERCRED not available"
+		have_remote_pid = false;
+#endif
+		/* Make sure sun_path is NULL terminated: */
+		sun = (struct sockaddr_un *)&osa.u.sa;
+		sun->sun_path[sizeof(sun->sun_path) - 1] = '\0';
+		if (have_remote_pid)
+			OSMO_STRBUF_PRINTF(sb, "r=%llu<->", remote_pid);
+		else
+			OSMO_STRBUF_PRINTF(sb, "r=NULL<->");
+		OSMO_STRBUF_PRINTF(sb, "l=%s:%d", sun->sun_path, fd);
+		return sb.chars_needed;
+	}
+	default:
+		osmo_strlcpy(str, "<socket-family-no-supported>", str_len);
+		return -ENOTSUP;
+	}
 }
 
 /*! Get address/port information on socket in static string, like "r=1.2.3.4:5<->l=6.7.8.9:10".
@@ -2222,6 +2721,83 @@ int osmo_sock_set_priority(int fd, int prio)
 	return setsockopt(fd, SOL_SOCKET, SO_PRIORITY, &prio, sizeof(prio));
 }
 
+#ifdef HAVE_LIBSCTP
+/*! Fill in array of struct sctp_paddrinfo with each of the remote addresses of an SCTP socket
+ *  \param[in] fd file descriptor of SCTP socket
+ *  \param[out] pinfo Pointer to memory holding an array of struct sctp_paddrinfo (pinfo_cnt length).
+ *  \param[in,out] pinfo_cnt length of pinfo array (in elements). On return it contains the number of addresses found.
+ *  \returns 0 on success; negative otherwise
+ *
+ * Upon return, pinfo_cnt can be set to a higher value than the one set by the
+ * caller. This can be used by the caller to find out the required array length
+ * and then obtaining by calling the function twice. Only up to pinfo_cnt addresses
+ * are filled in, as per the value provided by the caller.
+ *
+ * Usage example retrieving struct sctp_paddrinfo for all (up to OSMO_SOCK_MAX_ADDRS, 32) remote IP addresses:
+ * struct sctp_paddrinfo pinfo[OSMO_SOCK_MAX_ADDRS];
+ * size_t pinfo_cnt = ARRAY_SIZE(pinfo);
+ * rc = osmo_sock_sctp_get_peer_addr_info(fd, &pinfo[0], &num_hostbuf, pinfo_cnt);
+ * if (rc < 0)
+ *	goto error;
+ * if (pinfo_cnt > ARRAY_SIZE(hostbuf))
+ *	goto not_enough_buffers;
+ */
+int osmo_sock_sctp_get_peer_addr_info(int fd, struct sctp_paddrinfo *pinfo, size_t *pinfo_cnt)
+{
+	struct sockaddr *addrs = NULL;
+	unsigned int n_addrs, i;
+	void *addr_buf;
+	int rc;
+	socklen_t optlen;
+
+	rc = sctp_getpaddrs(fd, 0, &addrs);
+
+	if (rc < 0)
+		return rc;
+	if (rc == 0)
+		return -ENOTCONN;
+
+	n_addrs = rc;
+	addr_buf = (void *)addrs;
+	for (i = 0; i < n_addrs; i++) {
+		struct sockaddr *sa_addr = (struct sockaddr *)addr_buf;
+		size_t addrlen;
+
+		switch (sa_addr->sa_family) {
+		case AF_INET:
+			addrlen = sizeof(struct sockaddr_in);
+			break;
+		case AF_INET6:
+			addrlen = sizeof(struct sockaddr_in6);
+			break;
+		default:
+			rc = -EINVAL;
+			goto free_addrs_ret;
+		}
+
+		if (i >= *pinfo_cnt) {
+			addr_buf += addrlen;
+			continue;
+		}
+
+		memset(&pinfo[i], 0, sizeof(pinfo[0]));
+		memcpy(&pinfo[i].spinfo_address, sa_addr, addrlen);
+		optlen = sizeof(pinfo[0]);
+		rc = getsockopt(fd, SOL_SCTP, SCTP_GET_PEER_ADDR_INFO, &pinfo[i], &optlen);
+		if (rc < 0)
+			goto free_addrs_ret;
+
+		addr_buf += addrlen;
+	}
+
+	*pinfo_cnt = n_addrs;
+	rc = 0;
+free_addrs_ret:
+	sctp_freepaddrs(addrs);
+	return rc;
+}
+#endif
+
 #endif /* HAVE_SYS_SOCKET_H */
 
 /*! @} */
